import http
from typing import NamedTuple, Optional, Union, final

from souswift_core.utils import deprecated


class _Response(NamedTuple):
    message: str
    status_code: int


def _append_exclamation_mark(message: str):
    return message if message.endswith('!') else f'{message}!'


class APIError(Exception):
    _status = http.HTTPStatus.BAD_REQUEST

    def __init__(self, message: str) -> None:
        self._message = message

    @deprecated
    @classmethod
    def from_message(cls, message: str, status_code: Optional[int] = None):
        self = cls(message)
        if status_code:
            self._status = status_code
        return self

    @final
    def get_message(self):
        return _append_exclamation_mark(self._message)

    @final
    def response(self) -> _Response:
        return _Response(self.get_message(), self._status)


class DatabaseError(Exception):
    _status = http.HTTPStatus.BAD_REQUEST

    def __init__(self, target: str) -> None:
        self._target = target

    @deprecated
    @classmethod
    def from_object(cls, obj: Union[str, type]):
        if isinstance(obj, str):
            target = obj
        else:
            result_string = obj.__qualname__
            for item in obj.mro()[1:]:
                result_string = result_string.replace(item.__qualname__, '')
            target = result_string
        return cls(target)

    def get_message(self):
        return f'An error occured with {self._target}'

    @final
    def _get_message(self):
        return _append_exclamation_mark(self.get_message())

    def response(self) -> _Response:
        return _Response(self._get_message(), self._status)


class NotFoundError(DatabaseError):
    _status = http.HTTPStatus.NOT_FOUND

    def get_message(self):
        return f'{self._target} not found'


class ConflictError(DatabaseError):
    _status = http.HTTPStatus.CONFLICT

    def get_message(self):
        return f'{self._target} already exists'


class ForbiddenError(APIError):
    _status = http.HTTPStatus.FORBIDDEN

    def __init__(self) -> None:
        super().__init__('You do not have permission to use this route')


class InvalidPassword(ForbiddenError):
    def __init__(self) -> None:
        self._message = 'Invalid Password'


class UnsetPassword(APIError):
    def __init__(self) -> None:
        super().__init__('User has not set password yet')


class NotAuthenticated(APIError):
    _status = http.HTTPStatus.UNAUTHORIZED

    def __init__(self) -> None:
        self._message = 'not authenticated'


class InvalidOrExpiredToken(NotAuthenticated):
    def __init__(self) -> None:
        self._message = 'Invalid or expired token'



import logging
from abc import ABC, abstractmethod

from sqlalchemy import create_engine
from sqlalchemy.ext.asyncio import create_async_engine
from starlette.datastructures import State

from souswift_core.providers.config import DatabaseConfig


class DatabaseProvider(ABC):
    state_name = 'database_provider'

    def __init__(
        self, database_config: DatabaseConfig, logger: logging.Logger
    ) -> None:
        self.config = database_config
        self.logger = logger
        self.engine, self.sync_engine = self._create_engine()

    def _create_engine(self):
        return create_async_engine(
            self.config.get_uri(is_async=True), **self.config.pool_config
        ), create_engine(
            self.config.get_uri(is_async=False), **self.config.pool_config
        )

    @abstractmethod
    async def health_check(self):
        ...


def setup_database(
    provider_class: type[DatabaseProvider],
    config: DatabaseConfig,
    logger: logging.Logger,
):
    async def _setup_database(state: State):
        provider = provider_class(config, logger)
        setattr(
            state,
            DatabaseProvider.state_name,
            provider,
        )
        await provider.health_check()

    return _setup_database



from context_handler import context_factory
from context_handler._datastructures import AbstractAsyncContextFactory  # noqa
from sqlalchemy.ext.asyncio import AsyncConnection, AsyncSession

from .connection import ConnectionProvider
from .session import SessionProvider

session_factory: AbstractAsyncContextFactory[
    AsyncSession
] = context_factory(  # noqa
    SessionProvider, SessionProvider.Context
)
connection_factory: AbstractAsyncContextFactory[
    AsyncConnection
] = context_factory(  # noqa
    ConnectionProvider, ConnectionProvider.Context
)



import re
from typing import Generic, Optional, TypeVar

import fastapi
import pydantic
import pydantic.generics
import ujson

snake_regexp = re.compile(r'_([a-z])')


class Model(pydantic.BaseModel):
    class Config:
        allow_population_by_field_name = True
        json_loads = ujson.loads
        json_dumps = ujson.dumps

        @classmethod
        def alias_generator(cls, field: str):
            if field == 'id_':
                return 'id'
            return re.sub(snake_regexp, lambda match: match[1].upper(), field)


class FrozenModel(Model):
    class Config(Model.Config):
        allow_mutation = False


T = TypeVar('T', bound=Model)


class Details(FrozenModel):
    last_id: Optional[int] = None
    total_pages: Optional[int] = None
    info: Optional[str] = None
    total_left: Optional[int] = None


class BaseResponse(pydantic.generics.GenericModel, Generic[T]):
    data: list[T]
    details: Optional[Details] = pydantic.Field(default_factory=Details)


class Page(FrozenModel):
    last_id: int = fastapi.Query(0, ge=0)
    per_page: int = fastapi.Query(20, ge=5, le=100)


class BaseError(Model):
    title: str
    path: str
    status: int
    detail: str



import http
from typing import Union

import fastapi
import fastapi.responses

from souswift_core import dtos
from souswift_core.exc import exceptions

_Exception = Union[exceptions.APIError, exceptions.DatabaseError]


def _api_error_handler(request: fastapi.Request, exc: exceptions.APIError):
    exc_response = exc.response()
    return fastapi.responses.UJSONResponse(
        dtos.BaseError(
            title=type(exc).__name__.lower(),
            path=request.url.path,
            status=exc_response.status_code,
            detail=exc_response.message,
        ).dict(by_alias=True),
        status_code=exc_response.status_code,
    )


def _database_error_handler(
    request: fastapi.Request, exc: exceptions.DatabaseError
):
    exc_response = exc.response()
    return fastapi.responses.UJSONResponse(
        dtos.BaseError(
            title=type(exc).__name__.lower(),
            path=request.url.path,
            status=exc_response.status_code,
            detail=exc_response.message,
        ).dict(by_alias=True),
        status_code=exc_response.status_code,
    )


def _base_exception_handler(request: fastapi.Request, exc: Exception):
    status = http.HTTPStatus.INTERNAL_SERVER_ERROR
    return fastapi.responses.UJSONResponse(
        dtos.BaseError(
            title='exception',
            path=request.url.path,
            status=status,
            detail='Internal Server Error',
        ).dict(by_alias=True),
        status_code=status,
    )


def add_exception_handlers(
    app: fastapi.FastAPI,
    *exclude: type[_Exception],
    exclude_base_exception_handler: bool = False
):
    for exception, handler in (
        (exceptions.APIError, _api_error_handler),
        (exceptions.DatabaseError, _database_error_handler),
    ):
        if exception in exclude:
            continue
        app.add_exception_handler(exception, handler)
    if not exclude_base_exception_handler:
        app.add_exception_handler(Exception, _base_exception_handler)



import abc
import typing

from sqlalchemy import Boolean, Column, Table
from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql.elements import BooleanClauseList


class Filter:
    _mark: str = ''

    def where(self, entity: typing.Union[type, Table]):
        pass

    def __bool__(self):
        return False

    @property
    def mark(self):
        return self._mark

    def add_mark(self, mark: str):
        self._mark = mark
        return self


class Comparison(abc.ABC):
    @abc.abstractmethod
    def compare(
        self, attr: typing.Union[Column, typing.Any], target: typing.Any
    ) -> typing.Union[BooleanClauseList, 'ColumnElement[Boolean]']:
        """Receives column or function and returns a sqlalchemy comparison"""
